;; Generated package description from notmuch-indicator.el  -*- no-byte-compile: t -*-
(define-package "notmuch-indicator" "1.0.1.0.20231014.75027" "Display mode line indicator with notmuch-count(1) output" '((emacs "27.1")) :commit "8b03990994c04b604982ac672e7a01d1312950b3" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("notmuch-indicator Development" . "~protesilaos/general-issues@lists.sr.ht") :keywords '("convenience" "mail") :url "https://git.sr.ht/~protesilaos/notmuch-indicator")
